const { getJwks } = require('./utils/jwks.js');
const { jsonConfig } = require('./utils/jsonConfig.js');
const { validateKeys } = require('./utils/validate.js');
const jwt = require('jsonwebtoken');
const bodyParser = require('body-parser');
require('dotenv').config();
const forge = require('node-forge');
const fs = require('fs');
const { Sign, generateKey } = require('crypto');

let configuration = {};
let app; // This will hold an express app instance.

const config = (options) => {
    const requiredOptions = [
        'oidcInitiationUrl',
        'clientId',
        'baseURL',
        'launchPath',
        'ltiPath'
    ]
    const missingOptions = requiredOptions.filter(option => !(option in options));
    if (missingOptions.length > 0) {
        throw new Error(`Missing required options: ${missingOptions.join(', ')}`);
    }
    configuration = {...configuration, ...options};
}

const lti =  {
    config,
    expressInstance: function(expressInstance) {
        app = expressInstance;
        app.use(bodyParser.urlencoded({ extended: true }));

        const jwksPath = configuration.ltiPath ? `${configuration.ltiPath}/.well-known/jwks.json` : '/.well-known/jwks.json';
        app.get(jwksPath, async (req, res) => {
            try {
                const jwks = await getJwks();
                res.status(200).json(jwks);
            } catch (err) {
                console.error(err);
                res.status(500).json({ error: 'Unable to retrieve the JWKS for this application.' });
            }
        });

        const oidcPath = configuration.ltiPath ? `${configuration.ltiPath}/oidc` : '/oidc';
        app.post(oidcPath, async (req, res) => {
            const { login_hint, lti_message_hint, target_link_uri, client_id} = req.body;
            const oidcConfigUrl = configuration.oidcInitiationUrl;
            const redirectUri = target_link_uri;

            const oidcReq = {
                client_id: client_id,
                login_hint: login_hint,
                lti_message_hint: lti_message_hint,
                nonce: crypto.randomUUID(),
                prompt: 'none',
                redirect_uri: redirectUri,
                response_mode: 'form_post',
                response_type: 'id_token',
                scope: 'openid',
                state: crypto.randomUUID()
            };

            const queryString = Object.keys(oidcReq)
                .map(key => `${key}=${encodeURIComponent(oidcReq[key])}`)
                .join('&');
            
            const fullUrl = `${oidcConfigUrl}?${queryString}`;
            res.redirect(fullUrl);
        });
    },
    on: (path, handler) => {
        if (!configuration.ltiPath.startsWith('/')) configuration.ltiPath = `/${configuration.ltiPath}`;
        if (!configuration.ltiPath.endsWith('/')) configuration.ltiPath = `${configuration.ltiPath}/`;
        
        const fullPath = configuration.ltiPath ? `${configuration.ltiPath}${path}` : path;

        app.post(fullPath, async (req, res) => {
            try {
                const decoded = jwt.decode(req.body.id_token, { complete: true });
                if (decoded.payload['https://purl.imsglobal.org/spec/lti/claim/message_type'] === 'LtiDeepLinkingRequest') {
                    req.deepLinkingRequest = decoded.payload['https://purl.imsglobal.org/spec/lti-dl/claim/deep_linking_settings'].deep_link_return_url
                    req.deepLinkData = decoded.payload;
                    req.kid = decoded.header.kid;
                }
                await handler(req, res);
            } catch (err) {
                console.error(err);
                res.status(500).send('Error');
            }
        });
    },
    handleResponse: (req, res, items, options) => {
        const deepLinkReturnUrl = req.deepLinkingRequest;
        const contentItems = {
            "iss": configuration.clientId,
            "aud": 'https://canvas.instructure.com',
            "iat": Date.now()/1000,
            "exp": Date.now()/1000 + 60,
            "nonce": crypto.randomUUID(),
            "jti": crypto.randomUUID(),
            "sub": req.deepLinkData['sub'],
            "https://purl.imsglobal.org/spec/lti/claim/deployment_id": req.deepLinkData['https://purl.imsglobal.org/spec/lti/claim/deployment_id'],
            "https://purl.imsglobal.org/spec/lti/claim/message_type": 'LtiDeepLinkingResponse',
            "https://purl.imsglobal.org/spec/lti/claim/version": req.deepLinkData['https://purl.imsglobal.org/spec/lti/claim/version'],
            "https://purl.imsglobal.org/spec/lti-dl/claim/content_items": items,
            "https://purl.imsglobal.org/spec/lti-dl/claim/data": req.deepLinkData['https://purl.imsglobal.org/spec/lti-dl/claim/data'] ? req.deepLinkData['https://purl.imsglobal.org/spec/lti-dl/claim/data'] : '',
            "https://purl.imsglobal.org/spec/lti/claim/context": req.deepLinkData['https://purl.imsglobal.org/spec/lti/claim/context'],
        };
        // Assuming privateKey is in PEM format and not encrypted
        const privateKey = fs.readFileSync('private.key', 'utf8');
        // Sign the JWT
        const signedJwt = jwt.sign(contentItems, privateKey, {
            algorithm: 'RS256',
        });
        // Prepare the HTML response
        const htmlResponse = `
            <html>
                <body>
                    <form id="autoSubmit" action="${deepLinkReturnUrl}" method="POST">
                        <input type="hidden" name="JWT" value="${signedJwt}" />
                    </form>
                    <script>
                        window.onload = function() {
                            document.getElementById('autoSubmit').submit();
                        };
                    </script>
                </body>
            </html>
            `;

        // Send the response
        res.send(htmlResponse);
    },
    utils: {
        //jsonConfig,
        //validateKeys: validateKeys(),
        generateKeys: async function() {
            const jwks = await getJwks();
            if (jwks) {
                return true;
            } else {
                return false;
            }
        },
    }
    //jsonConfig,
}

module.exports = lti;